// #include <Arduino.h>
#include <KnxTpUart.h>
#include <Adafruit_GFX.h>    // Core graphics library
#include "Adafruit_EPD.h"
#include "Fonts/FreeSans9pt7b.h"
#include "Fonts/FreeSans12pt7b.h"
#include "Fonts/FreeMono9pt7b.h"
#include "Fonts/FreeMono12pt7b.h"
#include "Fonts/Org_01.h"


// Text Styles for Sizes
#define smallfont &FreeSans9pt7b
#define bigfont &FreeSans12pt7b

// Text colors / Change colorblack to "EPD_BLACK" if you use a multi colour display
// For some reason of the library the 1-colour display reacts to "EPD_RED" with black colour.
#define colorred EPD_RED
#define colorblack EPD_RED

// Text columns definition. Chance this if you want the columns to be different.
#define column1 10
#define column2 205


// EPD Pins Display 1
#define EPD_CS1     10
#define EPD_DC1     9
#define SRAM_CS1    8
#define EPD_RESET1   7

// EPD Pins Display 2 - if you use only one display, this can be removed. The rest of the pins are shared between the two displays.
#define EPD_CS2     6
#define EPD_DC2     5
#define SRAM_CS2    4
#define EPD_RESET2   3

// Address of this KNX device
#define PHYSICAL_ADDRESS "5.0.31"

// KNX Addresses definitions - Define your own addresses - Make sure the variables match in the code
// Use the replace function to replace all instances in the code. For example replace all "AWirkleistung" with 
// whatever variable you are using. For example "ATemperatureLivingRoom".  
// The names used here, are specific to my project. They make no sense anywhere else.

#define AWirkleistung "5/1/13"
#define AVerbrauch "5/1/22"
#define AWechselrichterleistung "5/0/11"
#define ASpeicherEnergie "2/7/12"
#define ASpeicherleistung "5/1/23"

#define AWaterCycle "5/0/19"            
#define ARegenwasser "5/0/8"
#define ATSolarRL "2/7/1"
#define ATSolarVL "2/7/0"
#define ATKesseltasche "2/7/11"
#define ATHeizkeller "2/5/16"

#define AHEsszimmerArdu "5/0/41"
#define APressure "5/0/42"
#define AeCO2 "5/0/44"

#define ATAussenSued "2/5/14"
#define ATAussenNord "2/5/13"
#define AHAussenNord "2/6/13"
#define ATWohnzimmer "2/5/4"
#define ATGaestezimmer "2/5/1"
#define ATVeranda "2/5/5"
#define ATSchlafzimmer "2/5/24"
#define ATSeminar "2/5/18"
#define ATAufenthaltsraum "2/5/28"
#define ATEstrich "2/5/11"

// Here Variables to display are organized in 4 groups, each group needs 5 arrays.
// The first array of each group is for the KNX address
// The second array of each group is the string that will be displayed on the display.
// The third array of each group is the unit of the displayed value
// The forth array of each group is the number of digits after the comma, that will be displayed.
// The fifth array of each group is the array to which the KNX Values will be stored after they are received.
// The order of the values has to be the same in each array.

// Please note: An array has a position 0 (Zero). It also has a length (number of places for values to be stored in)
// An array with this content: DEMO[] = {A, B, C, D, E} has a length of 5 (Five)
// BUT! The positions of the same array have numbers 0 - 4 (Four). 
// If you want to get the "E" from the array, which is the FIFTH value, you get it like this DEMO[4].
// This can sometimes be confusing. When you adapt this code you always need to be aware of this.
// In the comments of the code sometimes the array "position" is mentioned. Positions start with 0.

// Variable Group 1 (in this case, all variables that are about power)
char const *AdressVarLeis[] = {ASpeicherEnergie, ASpeicherleistung, AWechselrichterleistung, AVerbrauch, AWirkleistung};
char const *TextVarLeis[] = {"Speicherenergie:", "Speicherleistung:", "Photovoltaik:", "Stromverbrauch:", "Strombilanz:"};
char const *UnitVarLeis[] = {" kWh", " W", " W", " W", " W"};
int DigitVarLeis[] = {0, 0, 0, 0, 0};
volatile float ValueVarLeis[5]; // The number has to be the number of values that will be stored in here.


// Variable Group 2 (in this case, all variables that are about Temperatures of the heating system)
char const *AdressVarHeiz[] = {AWaterCycle, ARegenwasser, ATHeizkeller, ATSolarVL, ATSolarRL, ATKesseltasche};
char const *TextVarHeiz[] = {"R.Wasserbilanz", "Regenwasser:", "Heizkeller:", "Solarvorlauf:", "Solarruecklauf:", "Kesseltasche:"};
char const *UnitVarHeiz[] = {" L", " L", " C", " C", " C", " C"};
int const DigitVarHeiz[] = { 0, 0, 1, 1, 1, 1};
volatile float ValueVarHeiz[6];  // The number has to be the number of values that will be stored in here.


// Variable Group 3 (in this case, all variables that are sensing various stuff)
char const *AdressVarSens[] = {AHEsszimmerArdu, APressure, AeCO2};
char const *TextVarSens[] = {"Feuchte Innen:", "Luftdruck:", "CO2:"};
char const *UnitVarSens[] = {" %", " hPa", " ppm"};
int DigitVarSens[] = {0, 0, 0};
volatile float ValueVarSens[3];  // The number has to be the number of values that will be stored in here.


// Variable Group 4 (in this case, all variables that are about Temperatures of rooms)
char const *AdressVarTemp[] = {ATAussenSued, ATAussenNord, AHAussenNord, ATWohnzimmer, ATGaestezimmer, ATVeranda, ATSchlafzimmer, ATSeminar, ATAufenthaltsraum, ATEstrich};
char const *TextVarTemp[] = {"Aussen Sued:", "Aussen Nord:", "Feuchte aussen:", "Wohnzimmer:", "Gaestezimmer:", "Veranda:", "Schlafzimmer:", "Seminar:", "Aufenthaltsraum:", "Estrich:"};
char const *UnitVarTemp[] = {" C", " C", " %", " C", " C", " C", " C", " C", " C", " C"};
int DigitVarTemp[] = {1, 1, 0, 1, 1, 1, 1, 1, 1, 1};
volatile float ValueVarTemp[10];  // The number has to be the number of values that will be stored in here.



// The displays are initialized here. Other displays need changes here.

Adafruit_IL0398 display1(300, 400, EPD_DC1, EPD_RESET1, EPD_CS1, SRAM_CS1);
Adafruit_IL0398 display2(300, 400, EPD_DC2, EPD_RESET2, EPD_CS2, SRAM_CS2);

KnxTpUart knx(&Serial, PHYSICAL_ADDRESS);

// This is used in 2-colour displays to have a value show up in red when it is bellow 0. 
// Not used in this script. If you have a two colour display and define as such above, it will work on display two.
uint16_t condcolor;

unsigned long previousMillisEPD = 0; // Millis at which the EPD interval started
const long RefreshRate = 180; // Display Refresh Rate in seconds

void setup() {

  // Display config, Rotation of the displays is being set here.
  delay(5000);
  display1.begin();
  display1.setRotation(2);
  display1.clearBuffer();
  display1.display();
  delay(5000);
  display2.begin();
  display2.setRotation(2);
  display2.clearBuffer();
  display2.display();



  // KNX config
  Serial.begin(19200, SERIAL_8E1);
  knx.uartReset();


  // KNX Set listen addresses (Beware, original KnxTpUart.h, limits this to max 15. Edit KnxTpUart.h for more)
  // Each of the 4 groups are worked through here. The Numbers (4, 5, 9, 2) are according to ARRAYLENGTH - 1. 

  // KNX Set listen addresses for LeisVar  
  for (int i = 0; i <= 4; i = i + 1) {
    knx.addListenGroupAddress(AdressVarLeis[i]);
  }

  // KNX Set listen addresses for HeizVar
  for (int i = 0; i <= 5; i = i + 1) {
    knx.addListenGroupAddress(AdressVarHeiz[i]);
  }

  // KNX Set listen addresses for TempVar
  for (int i = 0; i <= 9; i = i + 1) {
    knx.addListenGroupAddress(AdressVarTemp[i]);

  }
  // KNX Set listen addresses for SensVar
  for (int i = 0; i <= 2; i = i + 1) {
    knx.addListenGroupAddress(AdressVarSens[i]);
  }
}


void loop() {

  // The KNX Read routine
  maintainKnxSerial();

  // Displays Write routine
  unsigned long currentMillisEPD = millis();
  if (currentMillisEPD - previousMillisEPD >= RefreshRate * 1000) {
    previousMillisEPD = currentMillisEPD;

    // Display 1 initialisation
    display1.clearBuffer();
    delay(1000); // Maybe needed for two displays not interfering... You can try without.
    display1.setTextSize(1); // Set text size
    display1.setTextColor(colorred); // this will be displayed black, its some weirdness in the library.
    // Draw some separation line on the display
    display1.drawLine(0, 127, display1.width(), 127, colorred);
    display1.drawLine(0, 128, display1.width(), 128, colorred);


    // The values are written to the display with the help of functions (WriteEPD1, WriteEPD2)
    // The structure for giving information to this function is: 
    // WriteEPD1(LineNumber to write  to, Text to be displayed, Value to be displayed, Unit to be displayed, How many digits to be displayed, If necessary an offset in pixels to the vertical position)

    // Here for display 1 the offsets are needed to correct the line positions for the separation line above.
    // The offsets to the vertical position for values before the separation line are negative to pull all lines up a bit
    // The offsets to the vertical position for values after the separation line are positive to push all lines down a bit
    
    // Write three values (position 0, 1, 2) of the temparature array to the display, 
    // start at line 1 of the display. 
    // Note: there is a position 0 in the array, but there is no Line 0 on the display. That is why its (i + 1) 
    
    for (int i = 0; i <= 2; i = i + 1) {
      WriteEPD1(i + 1 , TextVarTemp[i], ValueVarTemp[i], UnitVarTemp[i], DigitVarTemp[i], -3); // Temp + Humidity outside
    }
    
    // Write value stored in position 1 of the Sensor-Array to line 4 of the display.
    WriteEPD1(4, TextVarSens[1], ValueVarSens[1], UnitVarSens[1], DigitVarSens[1], -3); // Pressure
    
    // Write the values stored in positions 3 to 9 of the temperature array to the display starting at line 5 (i+2)
    for (int i = 3; i <= 9; i = i + 1) {
      WriteEPD1(i + 2 , TextVarTemp[i], ValueVarTemp[i], UnitVarTemp[i], DigitVarTemp[i], 5);
    }
    
    // Write value stored in position 0 of the Sensor-Array to line 12 of the display.
    WriteEPD1(12, TextVarSens[0], ValueVarSens[0], UnitVarSens[0], DigitVarSens[0], 5); // Humidity inside
    
    // Write value stored in position 2 of the Sensor-Array to line 13 of the display.
    WriteEPD1(13, TextVarSens[2], ValueVarSens[2], UnitVarSens[2], DigitVarSens[2], 5); // CO2 inside

    display1.display(); // End of display 1
    
    delay(5000); // The two displays might interfere if this is not here.

    // Display 2 
    display2.clearBuffer();
    display2.setTextSize(1);
    display2.setTextColor(colorblack);
    delay(1000); // The two displays might interfere if this is not here.

    // Write value stored in position 0 of the Power-Array to line 1 of the display.
    WriteEPD2(1, TextVarLeis[0], ValueVarLeis[0], UnitVarLeis[0], DigitVarLeis[0], 0); // Speicherenergie

    // Write the values stored in positions 1 to 4 of the Heating-array to the display starting at line 3 (i+2)
    for (int i = 1; i <= 4; i = i + 1) {
      WriteEPD2(i + 2, TextVarLeis[i], ValueVarLeis[i], UnitVarLeis[i], DigitVarLeis[i], 0); // Speicherleistung, PV, Stromverbrauch und Bilanz
    }

    // Write value stored in position 0 of the Heating-Array to line 7 of the display.
    WriteEPD2(7, TextVarHeiz[0], ValueVarHeiz[0], UnitVarHeiz[0], DigitVarHeiz[0], 0); // Regenwasserbilanz

    // Write value stored in position 1 of the Heating-Array to line 8 of the display.
    WriteEPD2(8, TextVarHeiz[1], ValueVarHeiz[1], UnitVarHeiz[1], DigitVarHeiz[1], 0); // Regenwasserstand

    // Write the values stored in positions 2 to 5 of the heating-array to the display starting at line 10 (i+8)
    for (int i = 2; i <= 5; i = i + 1) {
      WriteEPD2(i + 8, TextVarHeiz[i], ValueVarHeiz[i], UnitVarHeiz[i], DigitVarHeiz[i], 0); // Temperatures from heating system
    }

    // Draw a bar to visualize charge of heat storage, use line 2, Use The value in position 0 of the Power-Array, 650 represents the max value)
    drawhorizontalbar(2, ValueVarLeis[0], 650, colorred); 
    
    // Write the value in % into that bar 
    WriteEPD2Proz(2, ValueVarLeis[0], "%", 0, -2, 650); 

    // Draw a bar to visualize amount of rainwater, use line 9, Use The value in position 1 of the Power-Array, 10300 represents the max value)
    drawhorizontalbar(9, ValueVarHeiz[1], 10300, colorblack); // Draw a bar to visualize 
    WriteEPD2Proz(9, ValueVarHeiz[1], "%", 0, -2, 10300); // Write value in % into that bar

    display2.display(); // End of display 2

  } // End of Display write routine
} // End of Loop



// the KNX Read function 

void maintainKnxSerial() {
  if (Serial.available() > 0) {
    KnxTpUartSerialEventType eType = knx.serialEvent();
    if (eType == KNX_TELEGRAM) {
      KnxTelegram* telegram = knx.getReceivedTelegram();

      // Create the KNX String Address in the "0/0/0" format
      String target =
        String(0 + telegram->getTargetMainGroup()) + "/" +
        String(0 + telegram->getTargetMiddleGroup()) + "/" +
        String(0 + telegram->getTargetSubGroup());

      // Check what has been sent on KNX and write it into variables/arrays
      if (telegram->getCommand() == KNX_COMMAND_WRITE) {


        // Get the HeizGroup Vars from KNX - Needs adaptation to array size. Here it is 5 (6 positions)
        for (int i = 0; i <= 5; i = i + 1) {
          if (target == AdressVarHeiz[i]) {
            ValueVarHeiz[i] = telegram->get2ByteFloatValue();
          }
        }    
        // Get the SensGroup Vars from KNX  - Needs adaptation to array size. Here it is 2 (3 positions)
        for (int i = 0; i <= 2; i = i + 1) {
          if (target == AdressVarSens[i]) {
            ValueVarSens[i] = telegram->get2ByteFloatValue();
          }
        }
        // Get the TempGroup Vars from KNX  - Needs adaptation to array size. Here it is 9 (10 positions)
        for (int i = 0; i <= 9; i = i + 1) {
          if (target == AdressVarTemp[i]) {
            ValueVarTemp[i] = telegram->get2ByteFloatValue();
          }
        }

        
        // The LeisGroup is a bit complicated in this case. It is read in three steps. You probably don't need anyhting like this.
        
        // Get the LeisGroup Vars from KNX - Only the 3 positions are read here 0, 2, 3
        for (int i = 0; i <= 3; i = i + 1) {
          if (i == 1) continue; // Skip position 1
          if (target == AdressVarLeis[i]) {
            ValueVarLeis[i] = telegram->get2ByteFloatValue();
          }
        }
        // One LeisGroup Var (position 1) is read separately because it has to be multiplied by 1000 at this point.
        if (target == AdressVarLeis[1]) {
          ValueVarLeis[1] = telegram->get2ByteFloatValue() * 1000;
        }

        // One LeisGroup Var (position 4) is read separately because it is a 4Byte float and need rounding.
        if (target == AdressVarLeis[4]) {
          ValueVarLeis[4] = (-1) * round(telegram->get4ByteFloatValue());
        }


      }
    }
  }
}

// function to write to display 1
// Line Number, Text to display, Var to display, Unit to display

void WriteEPD1(int line, String Text, float fvalue, String unit, int digits, int offset) {
  display1.setFont(bigfont);
  display1.setCursor(column1, (lineheight * line)  + offset);
  display1.print(Text);
  display1.setCursor(column2, (lineheight * line)  + offset);
  display1.print(fvalue, digits);
  display1.setFont(smallfont);
  display1.print(unit);
}

// function to write to display 2, including conditional color for negative values
// Line Number, Text to display, Var to display, Unit to display

void WriteEPD2(int line, String Text, float Ivalue, String unit, int digits, int offset) {
  display2.setFont(bigfont);
  display2.setCursor(column1, (lineheight * line) - 2 + offset);
  display2.print(Text);
  display2.setCursor(column2, (lineheight * line) - 2 + offset);
  if (Ivalue >= 0) {
    condcolor = colorblack; // Values above 0 are black
  }
  else if (Ivalue < 0) {
    condcolor = colorred; // Values bellow 0 are red
  }
  display2.setTextColor(condcolor);
  display2.print(Ivalue, digits);
  display2.setFont(smallfont);
  display2.setTextColor(colorblack);
  display2.print(unit);

}

// function to write to display 2 in white color into the charge/fill bar,

void WriteEPD2Proz(int line, float Ivalue, String unit, int digits, int offset, int maximum) {
  display2.setFont(bigfont);
  display2.setTextColor(colorblack);
  display2.setFont(smallfont);
  int proz;
  proz = (int)round(Ivalue / maximum * 100);
  int columnproz;
  if (proz <= 85) {
    columnproz = column2 + 43;
    display2.setCursor(columnproz, (lineheight * line) - 2 + offset);
    display2.print(proz);
    display2.print(unit);
  }


}

// function for charge/fill bar. 
// Linenumber to be displayed on, value to be displayed, maximum the value can be for 100%, color

void drawhorizontalbar(int linenumber, float value, int maximum, uint16_t col) {

  // Specifiy charge-state bar horizontal
  int proz;
  proz = (int) round(value / maximum * 100);  int chargebarframeheight = lineheight - round(lineheight / 3); // The height of the chargebar frame, within upper group
  int chargebarframewidth = display2.width() - 2 * column1; // Width of chargebar
  int chargebarframestartX = column1;
  int chargebarframestartY = (linenumber - 1) * lineheight + lineheight / 3;

  // Draw Charge-state bar
  display2.drawRect(chargebarframestartX, chargebarframestartY, chargebarframewidth, chargebarframeheight, colorblack);
  int balken = round((proz  * (chargebarframewidth - 2) / 100 ));
  if (balken >= chargebarframewidth - 2) balken = chargebarframewidth - 2;
  display2.fillRect(chargebarframestartX + 1, chargebarframestartY + 1, balken, chargebarframeheight - 2, col);
}
